/*
 * linux/kernel/ptrace.c
 *
 * (C) Copyright 1999 Linus Torvalds
 *
 * Common interfaces for "ptrace()" which we do not want
 * to continually duplicate across every architecture.
 */

#include <linux/capability.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/pagemap.h>
#include <linux/smp_lock.h>
#include <linux/ptrace.h>
#include <linux/utrace.h>
#include <linux/security.h>
#include <linux/signal.h>
#include <linux/audit.h>
#include <linux/pid_namespace.h>
#include <linux/syscalls.h>

#include <asm/pgtable.h>
#include <asm/uaccess.h>

/*
 * ptrace a task: make the debugger its new parent and
 * move it to the ptrace list.
 *
 * Must be called with the tasklist lock write-held.
 */
void __ptrace_link(struct task_struct *child, struct task_struct *new_parent)
{
	BUG_ON(!list_empty(&child->ptrace_entry));
	list_add(&child->ptrace_entry, &new_parent->ptraced);
	child->parent = new_parent;
}
 
#ifndef CONFIG_UTRACE_PTRACE
/*
 * Turn a tracing stop into a normal stop now, since with no tracer there
 * would be no way to wake it up with SIGCONT or SIGKILL.  If there was a
 * signal sent that would resume the child, but didn't because it was in
 * TASK_TRACED, resume it now.
 * Requires that irqs be disabled.
 */
void ptrace_untrace(struct task_struct *child)
{
	spin_lock(&child->sighand->siglock);
	if (task_is_traced(child)) {
		if (child->signal->flags & SIGNAL_STOP_STOPPED) {
			__set_task_state(child, TASK_STOPPED);
		} else {
			signal_wake_up(child, 1);
		}
	}
	spin_unlock(&child->sighand->siglock);
}

static void ptrace_finish(struct task_struct *child)
{
	if (task_is_traced(child))
		ptrace_untrace(child);
}

static void ptrace_detach_task(struct task_struct *child)
{
	/* Architecture-specific hardware disable .. */
	ptrace_disable(child);
	clear_tsk_thread_flag(child, TIF_SYSCALL_TRACE);
}

static void utrace_engine_put(struct utrace_attached_engine *engine)
{
}

#else  /* CONFIG_UTRACE_PTRACE */

static const struct utrace_engine_ops ptrace_utrace_ops; /* forward decl */

static void ptrace_detach_task(struct task_struct *child)
{
	struct utrace_attached_engine *engine;
	engine = utrace_attach_task(child, UTRACE_ATTACH_MATCH_OPS,
				    &ptrace_utrace_ops, NULL);
	if (likely(!IS_ERR(engine))) {
		int ret = utrace_control(child, engine, UTRACE_DETACH);
		WARN_ON(ret && ret != -EINPROGRESS &&
			ret != -ESRCH && ret != -EALREADY);
		utrace_engine_put(engine);
	}
}

void ptrace_utrace_exit(struct task_struct *task)
{
	struct task_struct *child;
	read_lock(&tasklist_lock);
	list_for_each_entry(child, &task->ptraced, ptrace_entry)
		ptrace_detach_task(child);
	read_unlock(&tasklist_lock);
}

static void ptrace_finish(struct task_struct *child)
{
}

#endif	/* !CONFIG_UTRACE_PTRACE */

/*
 * unptrace a task: move it back to its original parent and
 * remove it from the ptrace list.
 *
 * Must be called with the tasklist lock write-held.
 */
void __ptrace_unlink(struct task_struct *child)
{
	BUG_ON(!child->ptrace);

	child->ptrace = 0;
	child->parent = child->real_parent;
	list_del_init(&child->ptrace_entry);

	ptrace_finish(child);
}

#ifndef CONFIG_UTRACE_PTRACE

/*
 * Check that we have indeed attached to the thing..
 */
int ptrace_check_attach(struct task_struct *child, int kill)
{
	int ret = -ESRCH;

	/*
	 * We take the read lock around doing both checks to close a
	 * possible race where someone else was tracing our child and
	 * detached between these two checks.  After this locked check,
	 * we are sure that this is our traced child and that can only
	 * be changed by us so it's not changing right after this.
	 */
	read_lock(&tasklist_lock);
	if ((child->ptrace & PT_PTRACED) && child->parent == current) {
		ret = 0;
		/*
		 * child->sighand can't be NULL, release_task()
		 * does ptrace_unlink() before __exit_signal().
		 */
		spin_lock_irq(&child->sighand->siglock);
		if (task_is_stopped(child))
			child->state = TASK_TRACED;
		else if (!task_is_traced(child) && !kill)
			ret = -ESRCH;
		spin_unlock_irq(&child->sighand->siglock);
	}
	read_unlock(&tasklist_lock);

	if (!ret && !kill)
		ret = wait_task_inactive(child, TASK_TRACED) ? 0 : -ESRCH;

	/* All systems go.. */
	return ret;
}

static struct utrace_attached_engine *ptrace_attach_utrace(
	struct task_struct *task)
{
	return NULL;
}

static void ptrace_detach_utrace(struct task_struct *task,
				 struct utrace_attached_engine *engine)
{
}

static int ptrace_update_utrace(struct task_struct *task,
				struct utrace_attached_engine *engine)
{
	return 0;
}

#else	/* CONFIG_UTRACE_PTRACE */

static int ptrace_update_utrace(struct task_struct *task,
				struct utrace_attached_engine *engine)
{
	unsigned long events;

	/*
	 * We need this for resume handling.
	 */
	events = UTRACE_EVENT(QUIESCE);

	/*
	 * These events are always reported.
	 */
	events |= UTRACE_EVENT(EXEC) | UTRACE_EVENT_SIGNAL_ALL;

	/*
	 * We always have to examine clone events to check for CLONE_PTRACE.
	 */
	events |= UTRACE_EVENT(CLONE);

	/*
	 * PTRACE_SETOPTIONS can request more events.
	 */
	if (task->ptrace & PT_TRACE_EXIT)
		events |= UTRACE_EVENT(EXIT);

	if (!engine) {
		int ret;
		engine = utrace_attach_task(task, UTRACE_ATTACH_MATCH_OPS,
					    &ptrace_utrace_ops, NULL);
		if (IS_ERR(engine))
			return -ESRCH;
		ret = utrace_set_events(task, engine, events);
		utrace_engine_put(engine);
		return ret;
	}

	return utrace_set_events(task, engine, events);
}

static int ptrace_unsafe_exec(struct utrace_attached_engine *engine,
			      struct task_struct *task)
{
	if (task->ptrace & PT_PTRACE_CAP)
		return LSM_UNSAFE_PTRACE_CAP;

	return LSM_UNSAFE_PTRACE;
}

static struct task_struct *ptrace_tracer_task(
	struct utrace_attached_engine *engine, struct task_struct *target)
{
	return target->parent;
}

static void ptrace_set_action(struct task_struct *task,
			      enum utrace_resume_action action,
			      enum utrace_syscall_action syscall)
{
	task->ptrace &= ~((UTRACE_SYSCALL_MASK | UTRACE_RESUME_MASK) << 16);
	task->ptrace |= ((UTRACE_RESUME - action) | syscall) << 16;
}

static enum utrace_resume_action ptrace_resume_action(struct task_struct *task)
{
	return UTRACE_RESUME - ((task->ptrace >> 16) & UTRACE_RESUME_MASK);
}

static enum utrace_syscall_action ptrace_syscall_action(
	struct task_struct *task)
{
	return (task->ptrace >> 16) & UTRACE_SYSCALL_MASK;
}

static u32 utrace_ptrace_report(u32 action, struct task_struct *task, int code)
{
	/*
	 * Special kludge magic in utrace.c (utrace_stop) sees this
	 * and calls do_notify_parent_cldstop() for us.  This kludge
	 * is necessary to keep that wakeup after we enter TASK_TRACED.
	 */
	ptrace_set_action(task, UTRACE_STOP, 0);

	task->exit_code = code;

	return action | UTRACE_STOP;
}

static u32 utrace_ptrace_event(struct task_struct *task,
			       int event, unsigned long msg)
{
	task->ptrace_message = msg;
	return utrace_ptrace_report(0, task, (event << 8) | SIGTRAP);
}

static u32 ptrace_report_exec(enum utrace_resume_action action,
			      struct utrace_attached_engine *engine,
			      struct task_struct *task,
			      const struct linux_binfmt *fmt,
			      const struct linux_binprm *bprm,
			      struct pt_regs *regs)
{
	if (task->ptrace & PT_TRACE_EXEC)
		return utrace_ptrace_event(task, PTRACE_EVENT_EXEC, 0);

	/*
	 * Old-fashioned ptrace'd exec just posts a plain signal.
	 */
	send_sig(SIGTRAP, task, 0);
	return UTRACE_RESUME;
}

static u32 ptrace_report_exit(enum utrace_resume_action action,
			      struct utrace_attached_engine *engine,
			      struct task_struct *task,
			      long orig_code, long *code)
{
	return utrace_ptrace_event(task, PTRACE_EVENT_EXIT, *code);
}

#define	PT_VFORKING	PT_DTRACE /* reuse obsolete bit */

static u32 ptrace_report_clone(enum utrace_resume_action action,
			       struct utrace_attached_engine *engine,
			       struct task_struct *parent,
			       unsigned long clone_flags,
			       struct task_struct *child)
{
	int event;
	struct utrace_attached_engine *child_engine;

	/*
	 * To simulate vfork-done tracing, we'll have to catch the
	 * parent's syscall-exit event for this vfork/clone system call.
	 * Since PTRACE_SETOPTIONS can enable PTRACE_O_TRACEVFORKDONE
	 * during the PTRACE_EVENT_VFORK stop, we must do this if either
	 * is enabled right now.
	 */
	if ((clone_flags & CLONE_VFORK) &&
	    (parent->ptrace & (PT_TRACE_VFORK | PT_TRACE_VFORK_DONE))) {
		if (!(engine->flags & UTRACE_EVENT(SYSCALL_EXIT))) {
			int ret = utrace_set_events(parent, engine,
						    engine->flags |
						    UTRACE_EVENT(SYSCALL_EXIT));
			WARN_ON(ret);
		}
		parent->ptrace |= PT_VFORKING;
	}

	if (clone_flags & CLONE_UNTRACED)
		return UTRACE_RESUME;

	event = 0;
	if (clone_flags & CLONE_VFORK) {
		if (parent->ptrace & PT_TRACE_VFORK)
			event = PTRACE_EVENT_VFORK;
	} else if ((clone_flags & CSIGNAL) != SIGCHLD) {
		if (parent->ptrace & PT_TRACE_CLONE)
			event = PTRACE_EVENT_CLONE;
	} else if (parent->ptrace & PT_TRACE_FORK) {
		event = PTRACE_EVENT_FORK;
	}

	/*
	 * Any of these reports implies auto-attaching the new child.
	 * So does CLONE_PTRACE, even with no event to report.
	 */
	if (event || (clone_flags & CLONE_PTRACE)) {
		child_engine = utrace_attach_task(child, UTRACE_ATTACH_CREATE |
						  UTRACE_ATTACH_EXCLUSIVE |
						  UTRACE_ATTACH_MATCH_OPS,
						  &ptrace_utrace_ops, NULL);
		if (unlikely(IS_ERR(child_engine))) {
			WARN_ON(1);	/* XXX */
		} else {
			BUG_ON(!child->ptrace);
			/* XXX already set by old ptrace code
			   task_lock(child);
			   child->ptrace = parent->ptrace;
			   child->parent = parent->parent;
			   task_unlock(child);
			*/
			ptrace_update_utrace(child, child_engine);
			utrace_engine_put(child_engine);
		}
	}

	if (event)
		return utrace_ptrace_event(parent, event, child->pid);

	return UTRACE_RESUME;
}


static u32 ptrace_report_syscall(u32 action, struct task_struct *task)
{
	int code = SIGTRAP;
	if (task->ptrace & PT_TRACESYSGOOD)
		code |= 0x80;
	return utrace_ptrace_report(action, task, code);
}

static u32 ptrace_report_syscall_entry(u32 action,
				       struct utrace_attached_engine *engine,
				       struct task_struct *task,
				       struct pt_regs *regs)
{
	/*
	 * If we're doing PTRACE_SYSEMU, just punt here and report
	 * at the exit stop instead.
	 */
	if (ptrace_syscall_action(task))
		return UTRACE_SYSCALL_ABORT | UTRACE_RESUME;

	return ptrace_report_syscall(UTRACE_SYSCALL_RUN, task);
}

static u32 ptrace_report_syscall_exit(enum utrace_resume_action action,
				      struct utrace_attached_engine *engine,
				      struct task_struct *task,
				      struct pt_regs *regs)
{
	if (!(engine->flags & UTRACE_EVENT(SYSCALL_ENTRY))) {
		/*
		 * We were not really using PTRACE_SYSCALL.
		 * SYSCALL_EXIT was only caught for vfork-done tracing.
		 */
		int ret = utrace_set_events(
			task, engine,
			engine->flags & ~UTRACE_EVENT(SYSCALL_EXIT));
		WARN_ON(ret);
		WARN_ON(!(task->ptrace & PT_VFORKING));
		task->ptrace &= ~PT_VFORKING;
		if (task->ptrace & PT_TRACE_VFORK_DONE)
			return utrace_ptrace_event(task,
						   PTRACE_EVENT_VFORK_DONE, 0);
		return UTRACE_RESUME;
	}

	if (task->ptrace & PT_VFORKING) {
		/*
		 * If we're reporting vfork-done, we'll have to
		 * remember to report syscall-exit after that.
		 */
		if (task->ptrace & PT_TRACE_VFORK_DONE)
			return utrace_ptrace_event(task,
						   PTRACE_EVENT_VFORK_DONE, 0);
		task->ptrace &= ~PT_VFORKING;
	}

	if (unlikely(ptrace_syscall_action(task)) &&
	    unlikely(ptrace_resume_action(task) == UTRACE_SINGLESTEP))
		/*
		 * This is PTRACE_SYSEMU_SINGLESTEP.
		 * Kludge: Prevent arch code from sending a SIGTRAP
		 * after tracehook_report_syscall_exit() returns.
		 */
		user_disable_single_step(task);

	return ptrace_report_syscall(0, task);
}

static u32 ptrace_resumed(struct task_struct *task, struct pt_regs *regs,
			  siginfo_t *info, struct k_sigaction *return_ka)
{
	/*
	 * This is not a new signal, but just a notification we
	 * asked for.  Either we're stopping after another report
	 * like exec or syscall, or we're resuming.
	 */
	if (ptrace_resume_action(task) == UTRACE_STOP)
		return UTRACE_SIGNAL_REPORT | UTRACE_STOP;

	/*
	 * We're resuming.  If there's no signal to deliver, just go.
	 * If we were given a signal, deliver it now.
	 */
	task->last_siginfo = NULL;
	if (!task->exit_code)
		return UTRACE_SIGNAL_REPORT | ptrace_resume_action(task);

	/* Update the siginfo structure if the signal has
	   changed.  If the debugger wanted something
	   specific in the siginfo structure then it should
	   have updated *info via PTRACE_SETSIGINFO.  */
	if (task->exit_code != info->si_signo) {
		info->si_signo = task->exit_code;
		info->si_errno = 0;
		info->si_code = SI_USER;
		info->si_pid = task_pid_vnr(task->parent);
		info->si_uid = task->parent->uid;
	}

	task->exit_code = 0;

	spin_lock_irq(&task->sighand->siglock);
	*return_ka = task->sighand->action[info->si_signo - 1];
	spin_unlock_irq(&task->sighand->siglock);

	return UTRACE_SIGNAL_DELIVER | ptrace_resume_action(task);
}

static u32 ptrace_report_signal(u32 action,
				struct utrace_attached_engine *engine,
				struct task_struct *task,
				struct pt_regs *regs,
				siginfo_t *info,
				const struct k_sigaction *orig_ka,
				struct k_sigaction *return_ka)
{
	/*
	 * Deal with a pending vfork-done event.  We'll stop again now
	 * for the syscall-exit report that was replaced with vfork-done.
	 */
	if (unlikely(task->ptrace & PT_VFORKING)) {
		task->ptrace &= ~PT_VFORKING;
		if ((engine->flags & UTRACE_EVENT(SYSCALL_ENTRY)) &&
		    utrace_signal_action(action) == UTRACE_SIGNAL_REPORT) {
			/*
			 * Make sure we get another report on wakeup.
			 */
			int x = utrace_control(task, engine, UTRACE_INTERRUPT);
			WARN_ON(x);
			return ptrace_report_syscall(UTRACE_SIGNAL_REPORT,
						     task);
		}
	}

	switch (utrace_signal_action(action)) {
	default:
		break;
	case UTRACE_SIGNAL_HANDLER:
		/*
		 * A handler was set up.  If we are stepping, pretend
		 * another SIGTRAP arrived.
		 */
		if (ptrace_resume_action(task) == UTRACE_SINGLESTEP ||
		    ptrace_resume_action(task) == UTRACE_BLOCKSTEP) {
			memset(info, 0, sizeof *info);
			info->si_signo = SIGTRAP;
			info->si_code = SIGTRAP;
			info->si_pid = task_pid_vnr(task);
			info->si_uid = task->uid;
			break;
		}
		/* Fall through.  */
	case UTRACE_SIGNAL_REPORT:
		return ptrace_resumed(task, regs, info, return_ka);
	}

	task->last_siginfo = info;
	return utrace_ptrace_report(UTRACE_SIGNAL_IGN, task, info->si_signo);
}

static u32 ptrace_report_quiesce(u32 action,
				 struct utrace_attached_engine *engine,
				 struct task_struct *task,
				 unsigned long event)
{
	/*
	 * Make sure we deal with a pending vfork-done event (see above).
	 */
	if (unlikely(task->ptrace & PT_VFORKING))
		return UTRACE_INTERRUPT;

	task->last_siginfo = NULL;
	return ptrace_resume_action(task);
}

static const struct utrace_engine_ops ptrace_utrace_ops = {
	.tracer_task = ptrace_tracer_task,
	.unsafe_exec = ptrace_unsafe_exec,
	.report_signal = ptrace_report_signal,
	.report_quiesce = ptrace_report_quiesce,
	.report_exec = ptrace_report_exec,
	.report_exit = ptrace_report_exit,
	.report_clone = ptrace_report_clone,
	.report_syscall_entry = ptrace_report_syscall_entry,
	.report_syscall_exit = ptrace_report_syscall_exit,
};

/*
 * Detach the utrace engine.
 */
static void ptrace_detach_utrace(struct task_struct *task,
				 struct utrace_attached_engine *engine)
{
	int ret = utrace_control(task, engine, UTRACE_DETACH);
	WARN_ON(ret && ret != -ESRCH && ret != -EALREADY);
}

/*
 * Attach a utrace engine for ptrace and set up its event mask.
 * Returns the engine pointer or an IS_ERR() pointer.
 */
static struct utrace_attached_engine *ptrace_attach_utrace(
	struct task_struct *child)
{
	struct utrace_attached_engine *engine;
	engine = utrace_attach_task(child, UTRACE_ATTACH_CREATE |
				    UTRACE_ATTACH_EXCLUSIVE |
				    UTRACE_ATTACH_MATCH_OPS,
				    &ptrace_utrace_ops, NULL);
	if (IS_ERR(engine))
		return engine;
	if (likely(!ptrace_update_utrace(child, engine)))
		return engine;
	ptrace_detach_utrace(child, engine);
	utrace_engine_put(engine);
	return ERR_PTR(-ESRCH);
}

int ptrace_check_attach(struct task_struct *child, int kill)
{
	struct utrace_attached_engine *engine;
	struct utrace_examiner exam;
	int ret;

	engine = utrace_attach_task(child, UTRACE_ATTACH_MATCH_OPS,
				    &ptrace_utrace_ops, NULL);
	if (IS_ERR(engine))
		return -ESRCH;

	/*
	 * Make sure our engine has already stopped the child.
	 * Then wait for it to be off the CPU.
	 */
	ret = 0;
	if (utrace_control(child, engine, UTRACE_STOP) ||
	    utrace_prepare_examine(child, engine, &exam))
		ret = -ESRCH;

	utrace_engine_put(engine);

	return ret;
}

#endif	/* !CONFIG_UTRACE_PTRACE */

int __ptrace_may_attach(struct task_struct *task)
{
	/* May we inspect the given task?
	 * This check is used both for attaching with ptrace
	 * and for allowing access to sensitive information in /proc.
	 *
	 * ptrace_attach denies several cases that /proc allows
	 * because setting up the necessary parent/child relationship
	 * or halting the specified task is impossible.
	 */
	int dumpable = 0;
	/* Don't let security modules deny introspection */
	if (task == current)
		return 0;
	if (((current->uid != task->euid) ||
	     (current->uid != task->suid) ||
	     (current->uid != task->uid) ||
	     (current->gid != task->egid) ||
	     (current->gid != task->sgid) ||
	     (current->gid != task->gid)) && !capable(CAP_SYS_PTRACE))
		return -EPERM;
	smp_rmb();
	if (task->mm)
		dumpable = get_dumpable(task->mm);
	if (!dumpable && !capable(CAP_SYS_PTRACE))
		return -EPERM;

	return security_ptrace(current, task);
}

int ptrace_may_attach(struct task_struct *task)
{
	int err;
	task_lock(task);
	err = __ptrace_may_attach(task);
	task_unlock(task);
	return !err;
}

int ptrace_attach(struct task_struct *task)
{
	int retval;
	unsigned long flags;
	struct utrace_attached_engine *engine;

	audit_ptrace(task);

	retval = -EPERM;
	if (same_thread_group(task, current))
		goto out;

	engine = ptrace_attach_utrace(task);
	if (unlikely(IS_ERR(engine))) {
		if (PTR_ERR(engine) == -ESRCH)
			retval = -ESRCH;
		goto out;
	}

repeat:
	/*
	 * Nasty, nasty.
	 *
	 * We want to hold both the task-lock and the
	 * tasklist_lock for writing at the same time.
	 * But that's against the rules (tasklist_lock
	 * is taken for reading by interrupts on other
	 * cpu's that may have task_lock).
	 */
	task_lock(task);
	if (!write_trylock_irqsave(&tasklist_lock, flags)) {
		task_unlock(task);
		do {
			cpu_relax();
		} while (!write_can_lock(&tasklist_lock));
		goto repeat;
	}

	if (!task->mm)
		goto bad;
	/* the same process cannot be attached many times */
	if (task->ptrace & PT_PTRACED)
		goto bad;
	retval = __ptrace_may_attach(task);
	if (retval)
		goto bad;

	/* Go */
	task->ptrace |= PT_PTRACED;
	if (capable(CAP_SYS_PTRACE))
		task->ptrace |= PT_PTRACE_CAP;

	__ptrace_link(task, current);

	send_sig_info(SIGSTOP, SEND_SIG_FORCED, task);
bad:
	write_unlock_irqrestore(&tasklist_lock, flags);
	task_unlock(task);
	if (!IS_ERR(engine)) {
		if (retval)
			ptrace_detach_utrace(task, engine);
		utrace_engine_put(engine);
	}
out:
	return retval;
}

static inline void __ptrace_detach(struct task_struct *child, unsigned int data)
{
	child->exit_code = data;
	/* .. re-parent .. */
	__ptrace_unlink(child);
	/* .. and wake it up. */
	if (child->exit_state != EXIT_ZOMBIE)
		wake_up_process(child);
}

int ptrace_detach(struct task_struct *child, unsigned int data)
{
	if (!valid_signal(data))
		return -EIO;

	ptrace_detach_task(child);

	write_lock_irq(&tasklist_lock);
	/* protect against de_thread()->release_task() */
	if (child->ptrace)
		__ptrace_detach(child, data);
	write_unlock_irq(&tasklist_lock);

	return 0;
}

int ptrace_readdata(struct task_struct *tsk, unsigned long src, char __user *dst, int len)
{
	int copied = 0;

	while (len > 0) {
		char buf[128];
		int this_len, retval;

		this_len = (len > sizeof(buf)) ? sizeof(buf) : len;
		retval = access_process_vm(tsk, src, buf, this_len, 0);
		if (!retval) {
			if (copied)
				break;
			return -EIO;
		}
		if (copy_to_user(dst, buf, retval))
			return -EFAULT;
		copied += retval;
		src += retval;
		dst += retval;
		len -= retval;			
	}
	return copied;
}

int ptrace_writedata(struct task_struct *tsk, char __user *src, unsigned long dst, int len)
{
	int copied = 0;

	while (len > 0) {
		char buf[128];
		int this_len, retval;

		this_len = (len > sizeof(buf)) ? sizeof(buf) : len;
		if (copy_from_user(buf, src, this_len))
			return -EFAULT;
		retval = access_process_vm(tsk, dst, buf, this_len, 1);
		if (!retval) {
			if (copied)
				break;
			return -EIO;
		}
		copied += retval;
		src += retval;
		dst += retval;
		len -= retval;			
	}
	return copied;
}

static int ptrace_setoptions(struct task_struct *child, long data)
{
	child->ptrace &= ~PT_TRACE_MASK;

	if (data & PTRACE_O_TRACESYSGOOD)
		child->ptrace |= PT_TRACESYSGOOD;

	if (data & PTRACE_O_TRACEFORK)
		child->ptrace |= PT_TRACE_FORK;

	if (data & PTRACE_O_TRACEVFORK)
		child->ptrace |= PT_TRACE_VFORK;

	if (data & PTRACE_O_TRACECLONE)
		child->ptrace |= PT_TRACE_CLONE;

	if (data & PTRACE_O_TRACEEXEC)
		child->ptrace |= PT_TRACE_EXEC;

	if (data & PTRACE_O_TRACEVFORKDONE)
		child->ptrace |= PT_TRACE_VFORK_DONE;

	if (data & PTRACE_O_TRACEEXIT)
		child->ptrace |= PT_TRACE_EXIT;

	ptrace_update_utrace(child, NULL);

	return (data & ~PTRACE_O_MASK) ? -EINVAL : 0;
}

static int ptrace_getsiginfo(struct task_struct *child, siginfo_t *info)
{
	int error = -ESRCH;

	read_lock(&tasklist_lock);
	if (likely(child->sighand != NULL)) {
		error = -EINVAL;
		spin_lock_irq(&child->sighand->siglock);
		if (likely(child->last_siginfo != NULL)) {
			*info = *child->last_siginfo;
			error = 0;
		}
		spin_unlock_irq(&child->sighand->siglock);
	}
	read_unlock(&tasklist_lock);
	return error;
}

static int ptrace_setsiginfo(struct task_struct *child, const siginfo_t *info)
{
	int error = -ESRCH;

	read_lock(&tasklist_lock);
	if (likely(child->sighand != NULL)) {
		error = -EINVAL;
		spin_lock_irq(&child->sighand->siglock);
		if (likely(child->last_siginfo != NULL)) {
			*child->last_siginfo = *info;
			error = 0;
		}
		spin_unlock_irq(&child->sighand->siglock);
	}
	read_unlock(&tasklist_lock);
	return error;
}


#ifdef PTRACE_SINGLESTEP
#define is_singlestep(request)		((request) == PTRACE_SINGLESTEP)
#else
#define is_singlestep(request)		0
#endif

#ifdef PTRACE_SINGLEBLOCK
#define is_singleblock(request)		((request) == PTRACE_SINGLEBLOCK)
#else
#define is_singleblock(request)		0
#endif

#ifdef PTRACE_SYSEMU
#define is_sysemu_singlestep(request)	((request) == PTRACE_SYSEMU_SINGLESTEP)
#else
#define is_sysemu_singlestep(request)	0
#endif

#ifndef CONFIG_UTRACE_PTRACE
static int ptrace_resume(struct task_struct *child, long request, long data)
{
	if (!valid_signal(data))
		return -EIO;

	if (request == PTRACE_SYSCALL)
		set_tsk_thread_flag(child, TIF_SYSCALL_TRACE);
	else
		clear_tsk_thread_flag(child, TIF_SYSCALL_TRACE);

#ifdef TIF_SYSCALL_EMU
	if (request == PTRACE_SYSEMU || request == PTRACE_SYSEMU_SINGLESTEP)
		set_tsk_thread_flag(child, TIF_SYSCALL_EMU);
	else
		clear_tsk_thread_flag(child, TIF_SYSCALL_EMU);
#endif

	if (is_singleblock(request)) {
		if (unlikely(!arch_has_block_step()))
			return -EIO;
		user_enable_block_step(child);
	} else if (is_singlestep(request) || is_sysemu_singlestep(request)) {
		if (unlikely(!arch_has_single_step()))
			return -EIO;
		user_enable_single_step(child);
	}
	else
		user_disable_single_step(child);

	child->exit_code = data;
	wake_up_process(child);

	return 0;
}
#else  /* CONFIG_UTRACE_PTRACE */
static int ptrace_resume(struct task_struct *child, long request, long data)
{
	struct utrace_attached_engine *engine;
	enum utrace_resume_action action;
	enum utrace_syscall_action syscall;
	int ret = 0;

	if (!valid_signal(data))
		return -EIO;

	engine = utrace_attach_task(child, UTRACE_ATTACH_MATCH_OPS,
				    &ptrace_utrace_ops, NULL);
	if (IS_ERR(engine))
		return -ESRCH;

	syscall = UTRACE_SYSCALL_RUN;
#ifdef PTRACE_SYSEMU
	if (request == PTRACE_SYSEMU || request == PTRACE_SYSEMU_SINGLESTEP)
		syscall = UTRACE_SYSCALL_ABORT;
#endif

	if (syscall != UTRACE_SYSCALL_RUN || request == PTRACE_SYSCALL) {
		if (!(engine->flags & UTRACE_EVENT_SYSCALL) &&
		    utrace_set_events(child, engine,
				      engine->flags | UTRACE_EVENT_SYSCALL))
			ret = -ESRCH;
	} else if (engine->flags & UTRACE_EVENT(SYSCALL_ENTRY)) {
		if (utrace_set_events(child, engine,
				      engine->flags & ~UTRACE_EVENT_SYSCALL))
			ret = -ESRCH;
	}

	action = UTRACE_RESUME;
	if (is_singleblock(request)) {
		if (unlikely(!arch_has_block_step()))
			ret = -EIO;
		action = UTRACE_BLOCKSTEP;
	} else if (is_singlestep(request) || is_sysemu_singlestep(request)) {
		if (unlikely(!arch_has_single_step()))
			ret = -EIO;
		action = UTRACE_SINGLESTEP;
	}

	if (!ret) {
		child->exit_code = data;

		ptrace_set_action(child, action, syscall);

		if (task_is_stopped(child)) {
			spin_lock_irq(&child->sighand->siglock);
			child->signal->flags &= ~SIGNAL_STOP_STOPPED;
			spin_unlock_irq(&child->sighand->siglock);
		}

		/*
		 * To resume with a signal we must hit ptrace_report_signal.
		 */
		if (data)
			action = UTRACE_INTERRUPT;

		if (utrace_control(child, engine, action))
			ret = -ESRCH;
	}

	utrace_engine_put(engine);

	return ret;
}
#endif	/* !CONFIG_UTRACE_PTRACE */

int ptrace_request(struct task_struct *child, long request,
		   long addr, long data)
{
	int ret = -EIO;
	siginfo_t siginfo;

	switch (request) {
	case PTRACE_PEEKTEXT:
	case PTRACE_PEEKDATA:
		return generic_ptrace_peekdata(child, addr, data);
	case PTRACE_POKETEXT:
	case PTRACE_POKEDATA:
		return generic_ptrace_pokedata(child, addr, data);

#ifdef PTRACE_OLDSETOPTIONS
	case PTRACE_OLDSETOPTIONS:
#endif
	case PTRACE_SETOPTIONS:
		ret = ptrace_setoptions(child, data);
		break;
	case PTRACE_GETEVENTMSG:
		ret = put_user(child->ptrace_message, (unsigned long __user *) data);
		break;

	case PTRACE_GETSIGINFO:
		ret = ptrace_getsiginfo(child, &siginfo);
		if (!ret)
			ret = copy_siginfo_to_user((siginfo_t __user *) data,
						   &siginfo);
		break;

	case PTRACE_SETSIGINFO:
		if (copy_from_user(&siginfo, (siginfo_t __user *) data,
				   sizeof siginfo))
			ret = -EFAULT;
		else
			ret = ptrace_setsiginfo(child, &siginfo);
		break;

	case PTRACE_DETACH:	 /* detach a process that was attached. */
		ret = ptrace_detach(child, data);
		break;

#ifdef PTRACE_SINGLESTEP
	case PTRACE_SINGLESTEP:
#endif
#ifdef PTRACE_SINGLEBLOCK
	case PTRACE_SINGLEBLOCK:
#endif
#ifdef PTRACE_SYSEMU
	case PTRACE_SYSEMU:
	case PTRACE_SYSEMU_SINGLESTEP:
#endif
	case PTRACE_SYSCALL:
	case PTRACE_CONT:
		return ptrace_resume(child, request, data);

	case PTRACE_KILL:
		if (child->exit_state)	/* already dead */
			return 0;
		return ptrace_resume(child, request, SIGKILL);

	default:
		break;
	}

	return ret;
}

/**
 * ptrace_traceme  --  helper for PTRACE_TRACEME
 *
 * Performs checks and sets PT_PTRACED.
 * Should be used by all ptrace implementations for PTRACE_TRACEME.
 */
int ptrace_traceme(void)
{
	int ret = -EPERM;
	struct utrace_attached_engine *engine;

	engine = ptrace_attach_utrace(current);
	if (unlikely(IS_ERR(engine)))
		return ret;

	/*
	 * Are we already being traced?
	 */
repeat:
	task_lock(current);
	if (!(current->ptrace & PT_PTRACED)) {
		/*
		 * See ptrace_attach() comments about the locking here.
		 */
		unsigned long flags;
		if (!write_trylock_irqsave(&tasklist_lock, flags)) {
			task_unlock(current);
			do {
				cpu_relax();
			} while (!write_can_lock(&tasklist_lock));
			goto repeat;
		}

		ret = security_ptrace(current->parent, current);

		/*
		 * Set the ptrace bit in the process ptrace flags.
		 * Then link us on our parent's ptraced list.
		 */
		if (!ret) {
			current->ptrace |= PT_PTRACED;
			__ptrace_link(current, current->real_parent);
		}

		write_unlock_irqrestore(&tasklist_lock, flags);
	}
	task_unlock(current);
	if (ret)
		ptrace_detach_utrace(current, engine);
	utrace_engine_put(engine);
	return ret;
}

/**
 * ptrace_get_task_struct  --  grab a task struct reference for ptrace
 * @pid:       process id to grab a task_struct reference of
 *
 * This function is a helper for ptrace implementations.  It checks
 * permissions and then grabs a task struct for use of the actual
 * ptrace implementation.
 *
 * Returns the task_struct for @pid or an ERR_PTR() on failure.
 */
struct task_struct *ptrace_get_task_struct(pid_t pid)
{
	struct task_struct *child;

	read_lock(&tasklist_lock);
	child = find_task_by_vpid(pid);
	if (child)
		get_task_struct(child);

	read_unlock(&tasklist_lock);
	if (!child)
		return ERR_PTR(-ESRCH);
	return child;
}

#ifndef arch_ptrace_attach
#define arch_ptrace_attach(child)	do { } while (0)
#endif

asmlinkage long sys_ptrace(long request, long pid, long addr, long data)
{
	struct task_struct *child;
	long ret;

	/*
	 * This lock_kernel fixes a subtle race with suid exec
	 */
	lock_kernel();
	if (request == PTRACE_TRACEME) {
		ret = ptrace_traceme();
		if (!ret)
			arch_ptrace_attach(current);
		goto out;
	}

	child = ptrace_get_task_struct(pid);
	if (IS_ERR(child)) {
		ret = PTR_ERR(child);
		goto out;
	}

	if (request == PTRACE_ATTACH) {
		ret = ptrace_attach(child);
		/*
		 * Some architectures need to do book-keeping after
		 * a ptrace attach.
		 */
		if (!ret)
			arch_ptrace_attach(child);
		goto out_put_task_struct;
	}

	ret = ptrace_check_attach(child, request == PTRACE_KILL);
	if (ret < 0)
		goto out_put_task_struct;

	ret = arch_ptrace(child, request, addr, data);
	if (ret < 0)
		goto out_put_task_struct;

 out_put_task_struct:
	put_task_struct(child);
 out:
	unlock_kernel();
	return ret;
}

int generic_ptrace_peekdata(struct task_struct *tsk, long addr, long data)
{
	unsigned long tmp;
	int copied;

	copied = access_process_vm(tsk, addr, &tmp, sizeof(tmp), 0);
	if (copied != sizeof(tmp))
		return -EIO;
	return put_user(tmp, (unsigned long __user *)data);
}

int generic_ptrace_pokedata(struct task_struct *tsk, long addr, long data)
{
	int copied;

	copied = access_process_vm(tsk, addr, &data, sizeof(data), 1);
	return (copied == sizeof(data)) ? 0 : -EIO;
}

#if defined CONFIG_COMPAT && defined __ARCH_WANT_COMPAT_SYS_PTRACE
#include <linux/compat.h>

int compat_ptrace_request(struct task_struct *child, compat_long_t request,
			  compat_ulong_t addr, compat_ulong_t data)
{
	compat_ulong_t __user *datap = compat_ptr(data);
	compat_ulong_t word;
	siginfo_t siginfo;
	int ret;

	switch (request) {
	case PTRACE_PEEKTEXT:
	case PTRACE_PEEKDATA:
		ret = access_process_vm(child, addr, &word, sizeof(word), 0);
		if (ret != sizeof(word))
			ret = -EIO;
		else
			ret = put_user(word, datap);
		break;

	case PTRACE_POKETEXT:
	case PTRACE_POKEDATA:
		ret = access_process_vm(child, addr, &data, sizeof(data), 1);
		ret = (ret != sizeof(data) ? -EIO : 0);
		break;

	case PTRACE_GETEVENTMSG:
		ret = put_user((compat_ulong_t) child->ptrace_message, datap);
		break;

	case PTRACE_GETSIGINFO:
		ret = ptrace_getsiginfo(child, &siginfo);
		if (!ret)
			ret = copy_siginfo_to_user32(
				(struct compat_siginfo __user *) datap,
				&siginfo);
		break;

	case PTRACE_SETSIGINFO:
		memset(&siginfo, 0, sizeof siginfo);
		if (copy_siginfo_from_user32(
			    &siginfo, (struct compat_siginfo __user *) datap))
			ret = -EFAULT;
		else
			ret = ptrace_setsiginfo(child, &siginfo);
		break;

	default:
		ret = ptrace_request(child, request, addr, data);
	}

	return ret;
}

asmlinkage long compat_sys_ptrace(compat_long_t request, compat_long_t pid,
				  compat_long_t addr, compat_long_t data)
{
	struct task_struct *child;
	long ret;

	/*
	 * This lock_kernel fixes a subtle race with suid exec
	 */
	lock_kernel();
	if (request == PTRACE_TRACEME) {
		ret = ptrace_traceme();
		goto out;
	}

	child = ptrace_get_task_struct(pid);
	if (IS_ERR(child)) {
		ret = PTR_ERR(child);
		goto out;
	}

	if (request == PTRACE_ATTACH) {
		ret = ptrace_attach(child);
		/*
		 * Some architectures need to do book-keeping after
		 * a ptrace attach.
		 */
		if (!ret)
			arch_ptrace_attach(child);
		goto out_put_task_struct;
	}

	ret = ptrace_check_attach(child, request == PTRACE_KILL);
	if (!ret)
		ret = compat_arch_ptrace(child, request, addr, data);

 out_put_task_struct:
	put_task_struct(child);
 out:
	unlock_kernel();
	return ret;
}
#endif	/* CONFIG_COMPAT && __ARCH_WANT_COMPAT_SYS_PTRACE */
